from typing import Literal, Mapping, Sequence, Any

from pydantic import BaseModel

from arena import models
from arena.models import Function, ChatCompletionToolParam, Message, ResponseFormat, TopLogprob, TokenLogprob, ChoiceLogprobs, Choice
"""
ChatCompletionCreate -> anthropic MessageCreateParams -> anthropic Message -> ChatCompletion
"""

MODELS = ("claude-3-opus-20240229", "claude-3-sonnet-20240229", "claude-3-haiku-20240307", "claude-2.1", "claude-2.0", "claude-instant-1.2")


class Metadata(BaseModel):
    user_id: str


class FunctionDefinition(BaseModel):
    name: str
    description: str | None = None
    input_schema: Mapping[str, Any]


class ChatCompletionRequest(BaseModel):
    """
    Maps to:
    https://github.com/anthropics/anthropic-sdk-python/blob/main/src/anthropic/types/message_create_params.py#L13
    https://docs.anthropic.com/claude/reference/messages_post
    """
    max_tokens: int = 1
    messages: Sequence[Message]
    model: str | Literal[*MODELS]
    metadata: Metadata | None = None
    stop_sequences: Sequence[str] | None = None
    system: str | None = None
    temperature: float | None = None
    tools: Sequence[FunctionDefinition] | None = None
    top_p: float | None = None
    top_k: int | None = None
    stream: bool | None = None

    @classmethod
    def from_chat_completion_request(cls, ccc: models.ChatCompletionRequest) -> "ChatCompletionRequest":
        messages: Sequence[Message] = [msg.model_dump() for msg in ccc.messages if not msg.role == "system"]
        system: Sequence[str] = [msg.content for msg in ccc.messages if msg.role == "system"]
        ccc = ccc.model_dump()
        if "max_tokens" in ccc:
            ccc["max_tokens"] = ccc["max_tokens"] or 1
        if "user" in ccc and ccc["user"] is not None:
            ccc["metadata"] = {"user_id": ccc["user"]}
            del ccc["user"]
        if "stop" in ccc:   
            ccc["stop_sequences"] = ccc["stop"]
            del ccc["stop"]
        if len(system)==0:
            ccc["system"] = None
        else:
            ccc["system"] = system[0]
        if "lm_config" in ccc:
            del ccc["lm_config"]
        ccc["messages"] = messages
        return ChatCompletionRequest.model_validate(ccc)

    def to_dict(self) -> Mapping[str, Any]:
        return self.model_dump(exclude_none=True)


class TextBlock(BaseModel):
    text: str = ""
    type: Literal["text"] = "text"

class CompletionUsage(BaseModel):
    input_tokens: int | None = None
    output_tokens: int | None = None


class ChatCompletionResponse(BaseModel):
    """
    https://github.com/anthropics/anthropic-sdk-python/blob/main/src/anthropic/types/message.py#L14
    """
    id: str
    content: Sequence[TextBlock]
    model: str
    role: Literal["assistant"] = "assistant"
    stop_reason: Literal["end_turn", "max_tokens", "stop_sequence"] | None = None
    stop_sequence: str | None = None
    type: Literal["message"] = "message"
    usage: CompletionUsage | None = None

    @classmethod
    def from_dict(cls, m: Mapping[str, Any]) -> "ChatCompletionResponse":
        return ChatCompletionResponse.model_validate(m)

    def to_chat_completion_response(self) -> models.ChatCompletionResponse:
        finish_reasons = {
            "end_turn": "tool_calls",
            "max_tokens": "length",
            "stop_sequence": "stop"
        }
        return models.ChatCompletionResponse(
            id=self.id,
            choices=[Choice(
                finish_reason=finish_reasons.get(self.stop_reason, None),
                index=i,
                logprobs=None,
                message=Message(role="assistant", content=cb.text)
            ) for i,cb in enumerate(self.content)],
            model=self.model,
            usage=models.CompletionUsage(
                prompt_tokens=self.usage.input_tokens,
                completion_tokens=self.usage.output_tokens,
                total_tokens=self.usage.input_tokens + self.usage.output_tokens,
            ),
        )



from typing import Literal
from pydantic import BaseModel

# A LM config setting
class LMConfig(BaseModel):
    pii_removal: Literal["masking", "replace"] | None = None
    judge_evaluation: bool = False



from typing import Literal, Mapping, Sequence, Any
from pydantic import BaseModel
from arena.models.settings import LMConfig


"""All LanguageModels"""

class LMApiKeys(BaseModel):
    openai_api_key: str
    mistral_api_key: str
    anthropic_api_key: str

"""ChatCompletionCreate"""

# Inspired by https://github.com/mistralai/client-python/tree/main/src/mistralai/models
# And: https://github.com/anthropics/anthropic-sdk-python/tree/main/src/anthropic/types
# And: https://github.com/openai/openai-python/tree/main/src/openai/types


class Function(BaseModel):
    arguments: str
    name: str


class FunctionDefinition(BaseModel):
    name: str
    description: str | None = None
    parameters: Mapping[str, Any]


class ChatCompletionToolParam(BaseModel):
    id: str | None = None
    function: Function
    type: Literal["function"]


class Message(BaseModel):
    """
    Maps to:
    https://github.com/openai/openai-python/blob/main/src/openai/types/chat/chat_completion_message_param.py#L15
    https://github.com/openai/openai-python/blob/main/src/openai/types/chat/chat_completion_user_message_param.py#L13
    https://github.com/anthropics/anthropic-sdk-python/blob/main/src/anthropic/types/message_param.py#L15
    """
    content: str
    role: Literal["system", "user", "assistant", "tools"]
    name: str | None = None
    tool_calls: Sequence[ChatCompletionToolParam] | None = None
    """The tool calls generated by the model, such as function calls."""


class ResponseFormat(BaseModel):
    type: Literal["text", "json_object"] | None = None


class ChatCompletionRequest(BaseModel):
    """
    Maps to:
    https://github.com/openai/openai-python/blob/main/src/openai/types/chat/completion_create_params.py#L24
    https://github.com/mistralai/client-python/blob/main/src/mistralai/client.py#L153
    https://github.com/mistralai/client-python/blob/main/src/mistralai/models/chat_completion.py
    https://github.com/anthropics/anthropic-sdk-python/blob/main/src/anthropic/types/message_create_params.py#L13
    """
    messages: Sequence[Message]
    model: str | None
    frequency_penalty: float | None = None
    logit_bias: Mapping[str, int] | None = None
    logprobs: bool | None = None
    max_tokens: int | None = None
    n: int | None = None
    presence_penalty: float | None = None
    response_format: ResponseFormat | None = None
    safe_prompt: bool | None = None
    seed: int | None = None
    stop: str | Sequence[str] | None = None
    temperature: float | None = None
    tool_choice: Literal["none", "auto"] | ChatCompletionToolParam | None = None
    tools: Sequence[ChatCompletionToolParam] | None = None
    top_logprobs: int | None = None
    top_p: float | None = None
    user: str | None = None
    stream: bool | None = None
    lm_config: LMConfig | None = None

    def to_dict(self) -> Mapping[str, Any]:
        return self.model_dump(exclude_none=True)


"""ChatCompletion"""

class TopLogprob(BaseModel):
    token: str
    bytes: Sequence[int] | None = None
    logprob: float


class TokenLogprob(BaseModel):
    token: str
    logprob: float
    top_logprobs: Sequence[TopLogprob]


class ChoiceLogprobs(BaseModel):
    content: Sequence[TokenLogprob] | None = None


class Choice(BaseModel):    
    finish_reason: Literal["stop", "length", "tool_calls", "content_filter", "function_call", "error"] | None = None
    index: int
    logprobs: ChoiceLogprobs | None = None
    message: Message


class CompletionUsage(BaseModel):
    completion_tokens: int | None = None
    prompt_tokens: int
    total_tokens: int


class ChatCompletionResponse(BaseModel):
    """
    Maps to:
    https://github.com/openai/openai-python/blob/main/src/openai/types/chat/chat_completion.py#L40
    https://github.com/mistralai/client-python/blob/main/src/mistralai/models/chat_completion.py#L86
    https://github.com/anthropics/anthropic-sdk-python/blob/main/src/anthropic/types/message.py#L14
    """
    id: str
    choices: Sequence[Choice]
    created: int | None = None
    model: str
    object: Literal["chat.completion"] | None = None
    system_fingerprint: str | None = None
    usage: CompletionUsage | None = None
    lm_config: LMConfig | None = None

    @classmethod
    def from_dict(cls, m: Mapping[str, Any]) -> "ChatCompletionResponse":
        return ChatCompletionResponse.model_validate(m)



from typing import Any, Type, Literal
import httpx
import mistralai.client
from arena.models import ChatCompletionRequest, ChatCompletionResponse, Evaluation, Score, LMConfig, EventOut, ChatCompletionRequestEventResponse
import arena.models.openai as oai
import arena.models.mistral as mis
import arena.models.anthropic as ant
import openai
import mistralai
import anthropic

BASE_URL = "https://arena.sarus.app/api/v1"

class Client:
    def __init__(self, username: str | None = None, password: str | None = None, api_key: str | None = None, base_url: str = BASE_URL):
        self.username = username
        self.password = password
        self.api_key = api_key
        self.base_url = base_url
        self.timeout = 30.0
        if not self.api_key:
            self.login()
    
    
    def login(self):
        """Used internally to get a connection token from a user and a password"""
        assert(self.username and self.password)
        with httpx.Client(timeout=self.timeout) as client:
            resp = client.post(
                url = f"{self.base_url}/login/access-token",
                headers =  {
                    "accept": "application/json",
                    "Content-Type": "application/x-www-form-urlencoded",
                },
                data = {
                    "grant_type": "",
                    "username": self.username,
                    "password": self.password,
                    "scope": "",
                    "client_id": "",
                    "client_secret": "",
                },
            )
        self.api_key = resp.json()['access_token']
    

    def user(self, email: str, password: str, full_name: str | None = None):
        """Create a new user"""
        with httpx.Client(timeout=self.timeout) as client:
            client.post(
                url = f"{self.base_url}/users/open",
                headers = {
                    "Authorization": f"Bearer {self.api_key}"
                },
                json={
                    "email": email,
                    "password": password,
                    "full_name": full_name
                },
            )

    @staticmethod
    def user_open(email: str, password: str, full_name: str | None = None,  base_url=BASE_URL):
        """Create a new user"""
        with httpx.Client() as client:
            client.post(
                url = f"{base_url}/users/open",
                json={
                    "email": email,
                    "password": password,
                    "full_name": full_name
                },
            )


    def openai_api_key(self, api_key: str):
        """Set the OpenAI API token"""
        with httpx.Client(timeout=self.timeout) as client:
            client.post(
                url = f"{self.base_url}/settings/",
                headers = {
                    "Authorization": f"Bearer {self.api_key}"
                },
                json={
                    "name": "OPENAI_API_KEY",
                    "content": api_key
                },
            )
    

    def mistral_api_key(self, api_key: str):
        """Set the Mistral API token"""
        with httpx.Client(timeout=self.timeout) as client:
            client.post(
                url = f"{self.base_url}/settings/",
                headers = {
                    "Authorization": f"Bearer {self.api_key}"
                },
                json={
                    "name": "MISTRAL_API_KEY",
                    "content": api_key
                },
            )
    

    def anthropic_api_key(self, api_key: str):
        """Set the Anthropic API token"""
        with httpx.Client(timeout=self.timeout) as client:
            client.post(
                url = f"{self.base_url}/settings/",
                headers = {
                    "Authorization": f"Bearer {self.api_key}"
                },
                json={
                    "name": "ANTHROPIC_API_KEY",
                    "content": api_key
                },
            )
    

    def api_keys(self, openai_api_key: str, mistral_api_key: str, anthropic_api_key: str):
        """Set all API tokens"""
        self.openai_api_key(openai_api_key)
        self.mistral_api_key(mistral_api_key)
        self.anthropic_api_key(anthropic_api_key)


    def lm_config(self, lm_config: LMConfig):
        """Set LM config"""
        with httpx.Client(timeout=self.timeout) as client:
            client.post(
                url = f"{self.base_url}/settings/",
                headers = {
                    "Authorization": f"Bearer {self.api_key}"
                },
                json={
                    "name": "LM_CONFIG",
                    "content": lm_config.model_dump_json()
                },
            )


    def chat_completions(self, **kwargs: Any) -> ChatCompletionResponse:
        """Abstract chat completion"""
        req = ChatCompletionRequest.model_validate(kwargs).model_dump(mode="json", exclude_none=True)
        with httpx.Client(timeout=self.timeout) as client:
            resp = client.post(
                url = f"{self.base_url}/lm/chat/completions",
                headers = {
                    "Authorization": f"Bearer {self.api_key}"
                },
                json=req,
            )
        if resp.status_code == 200:
            return ChatCompletionResponse.model_validate(resp.json())
        else:
            raise RuntimeError(resp)
    
    def chat_completions_request(self, **kwargs: Any) -> EventOut:
        req = ChatCompletionRequest.model_validate(kwargs).model_dump(mode="json", exclude_none=True)
        with httpx.Client(timeout=self.timeout) as client:
            resp = client.post(
                url = f"{self.base_url}/lm/chat/completions/request",
                headers = {
                    "Authorization": f"Bearer {self.api_key}"
                },
                json=req,
            )
        if resp.status_code == 200:
            return EventOut.model_validate(resp.json())
        else:
            raise RuntimeError(resp)
    

    def chat_completions_response(self, **kwargs: Any) -> EventOut:
        req = ChatCompletionRequestEventResponse.model_validate(kwargs).model_dump(mode="json", exclude_none=True)
        with httpx.Client(timeout=self.timeout) as client:
            resp = client.post(
                url = f"{self.base_url}/lm/chat/completions/response",
                headers = {
                    "Authorization": f"Bearer {self.api_key}"
                },
                json=req,
            )
        if resp.status_code == 200:
            return EventOut.model_validate(resp.json())
        else:
            raise RuntimeError(resp)


    def evaluation(self, identifier: str, score: float) -> int:
        """Evaluate the response"""
        req = Evaluation(identifier=identifier, value=Score(value=score)).model_dump(mode="json", exclude_none=True)
        with httpx.Client(timeout=self.timeout) as client:
            resp = client.post(
                url = f"{self.base_url}/lm/evaluation",
                headers = {
                    "Authorization": f"Bearer {self.api_key}"
                },
                json=req,
            )
        if resp.status_code == 200:
            return resp.json()["id"]
        else:
            raise RuntimeError(resp)
    

    def decorate(self, client: Type, mode: Literal['proxy', 'instrument'] = 'proxy'):
        """Decorate the client"""
        if not hasattr(client, "_arena_decorated_"):
            client._arena_decorated_ = mode
            if client == openai.OpenAI:
                if mode == 'proxy':
                    self.decorate_openai_proxy(client)
                elif mode == 'instrument':
                    self.decorate_openai_instrument(client)
            elif client == mistralai.client.MistralClient:
                if mode == 'proxy':
                    self.decorate_mistral_proxy(client)
                elif mode == 'instrument':
                    self.decorate_mistral_instrument(client)
            elif client == anthropic.Anthropic:
                if mode == 'proxy':
                    self.decorate_anthropic_proxy(client)
                elif mode == 'instrument':
                    self.decorate_anthropic_instrument(client)


    def decorate_openai_proxy(self, client: Type[openai.OpenAI]):
        """Decorate openai client in proxy mode"""
        arena = self
        openai_init = client.__init__
        def init(self, *args: Any, **kwargs: Any):
            openai_init(self, *args, **kwargs)
            arena.openai_api_key(self.api_key)
            self.api_key = arena.api_key
            self.base_url = f"{arena.base_url}/lm/openai"
        client.__init__ = init
    

    def decorate_mistral_proxy(self, client: Type[mistralai.client.MistralClient]):
        """Decorate mistral client in proxy mode"""
        arena = self
        mistral_init = client.__init__
        def init(self, *args: Any, **kwargs: Any):
            mistral_init(self, *args, **kwargs)
            arena.mistral_api_key(self.api_key)
            self.api_key = arena.api_key
            self.base_url = f"{arena.base_url}/lm/mistral"
        client.__init__ = init
    

    def decorate_anthropic_proxy(self, client: Type[anthropic.Anthropic]):
        """Decorate anthropic client in proxy mode"""
        arena = self
        anthropic_init = client.__init__
        def init(self, *args: Any, **kwargs: Any):
            anthropic_init(self, *args, **kwargs)
            arena.anthropic_api_key(self.api_key)
            self.api_key = arena.api_key
            self.base_url = f"{arena.base_url}/lm/anthropic"
        client.__init__ = init
    

    def decorate_openai_instrument(self, client: Type[openai.OpenAI]):
        """Decorate openai client in instrument mode"""
        arena = self
        openai_init = client.__init__
        def init(self, *args: Any, **kwargs: Any):
            openai_init(self, *args, **kwargs)
            instance_create = self.chat.completions.create
            def chat_completion(*args: Any, **kwargs: Any):
                result = instance_create(*args, **kwargs)
                request = oai.ChatCompletionRequest(*args, **kwargs)
                request_event = arena.chat_completions_request(**request.model_dump(mode="json", exclude_none=True))
                response = oai.ChatCompletionResponse.model_validate(result.model_dump())
                arena.chat_completions_response(request=request, request_event_id=request_event.id, response=response)
                return result
            self.chat.completions.create = chat_completion
        client.__init__ = init
    

    def decorate_mistral_instrument(self, client: Type[mistralai.client.MistralClient]):
        """Decorate mistral client in instrument mode"""
        arena = self
        mistral_init = client.__init__
        def init(self, *args: Any, **kwargs: Any):
            mistral_init(self, *args, **kwargs)
            instance_chat = self.chat
            def chat_completion(*args: Any, **kwargs: Any):
                result = instance_chat(*args, **kwargs)
                request = mis.ChatCompletionRequest(*args, **kwargs)
                request_event = arena.chat_completions_request(**request.model_dump(mode="json", exclude_none=True))
                response = mis.ChatCompletionResponse.model_validate_json(result.model_dump_json())
                arena.chat_completions_response(request=request, request_event_id=request_event.id, response=response)
                return result
            self.chat = chat_completion
        client.__init__ = init
    

    def decorate_anthropic_instrument(self, client: Type[anthropic.Anthropic]):
        """Decorate anthropic client in instrument mode"""
        arena = self
        anthropic_init = client.__init__
        def init(self, *args: Any, **kwargs: Any):
            anthropic_init(self, *args, **kwargs)
            instance_create = self.messages.create
            def chat_completion(*args: Any, **kwargs: Any):
                result = instance_create(*args, **kwargs)
                request = ant.ChatCompletionRequest(*args, **kwargs)
                request_event = arena.chat_completions_request(**request.model_dump(mode="json", exclude_none=True))
                response = ant.ChatCompletionResponse.model_validate(result.model_dump())
                arena.chat_completions_response(request=request, request_event_id=request_event.id, response=response)
                return result
            self.messages.create = chat_completion
        client.__init__ = init



from typing import Mapping, Sequence, Literal, Any

from pydantic import BaseModel

from arena import models
from arena.models import Function, FunctionDefinition, ChatCompletionToolParam, Message, ResponseFormat, TopLogprob, TokenLogprob, ChoiceLogprobs, Choice, CompletionUsage

MODELS = ("gpt-4o", "gpt-4-turbo", "gpt-4", "gpt-3.5-turbo", "gpt-3.5-turbo-16k")

class ChatCompletionRequest(models.ChatCompletionRequest):
    """
    https://github.com/openai/openai-python/blob/main/src/openai/types/chat/completion_create_params.py#L24
    https://platform.openai.com/docs/api-reference/chat
    """
    model: str | Literal[*MODELS]

    @classmethod
    def from_chat_completion_request(cls, ccc: models.ChatCompletionRequest) -> "ChatCompletionRequest":
        return ChatCompletionRequest.model_validate(ccc.model_dump(exclude=["lm_config"]))

    def to_dict(self) -> Mapping[str, Any]:
        return self.model_dump(exclude_none=True)

class ChatCompletionResponse(models.ChatCompletionResponse):
    """
    https://github.com/openai/openai-python/blob/main/src/openai/types/chat/chat_completion.py#L40
    """
    @classmethod
    def from_dict(cls, m: Mapping[str, Any]) -> "ChatCompletionResponse":
        return ChatCompletionResponse.model_validate(m)

    def to_chat_completion_response(self) -> models.ChatCompletionResponse:
        return models.ChatCompletionResponse.model_validate(self.model_dump(exclude_none=True))




from datetime import datetime
from pydantic import BaseModel
from arena.models.chat_completion import (
    LMApiKeys,
    Function, FunctionDefinition,
    ChatCompletionToolParam, Message, ResponseFormat, ChatCompletionRequest,
    TopLogprob, TokenLogprob, ChoiceLogprobs, Choice, CompletionUsage, ChatCompletionResponse,
    )
from arena.models.evaluation import Evaluation, Score
from arena.models.settings import LMConfig
import arena.models.openai as oai
import arena.models.mistral as mis
import arena.models.anthropic as ant


class EventOut(BaseModel):
    id: int
    timestamp: datetime
    name: str
    content: str
    owner_id: int
    parent_id: int | None = None


class ChatCompletionRequestEventResponse(BaseModel):
    request: ChatCompletionRequest | oai.ChatCompletionRequest | mis.ChatCompletionRequest | ant.ChatCompletionRequest | None = None
    request_event_id: int
    response: ChatCompletionResponse | oai.ChatCompletionResponse | mis.ChatCompletionResponse | ant.ChatCompletionResponse


from typing import Literal, Mapping, Sequence, Any

from pydantic import BaseModel

from arena import models
from arena.models import Function, FunctionDefinition, ChatCompletionToolParam, Message, ResponseFormat, TopLogprob, TokenLogprob, ChoiceLogprobs, Choice, CompletionUsage

"""
models.ChatCompletionCreate -> ChatCompletionCreate -> ChatCompletion -> models.ChatCompletion
"""

MODELS = ("mistral-large-latest", "mistral-medium", "mistral-medium-latest", "mistral-small", "mistral-small-latest", "mistral-tiny", "open-mistral-7b", "open-mixtral-8x7b")


class ChatCompletionRequest(BaseModel):
    """
    Maps to:
    https://github.com/mistralai/client-python/blob/main/src/mistralai/client.py#L153
    https://github.com/mistralai/client-python/blob/main/src/mistralai/models/chat_completion.py
    https://docs.mistral.ai/api/#operation/createChatCompletion
    """
    messages: Sequence[Message]
    model: str | Literal[*MODELS]
    max_tokens: int | None = None
    response_format: ResponseFormat | None = None
    safe_prompt: bool | None = None
    random_seed: int | None = None
    temperature: float | None = None
    tool_choice: Literal["none", "auto", "any"] | None = None
    tools: Sequence[ChatCompletionToolParam] | None = None
    top_p: float | None = None
    stream: bool | None = None

    @classmethod
    def from_chat_completion_request(cls, ccc: models.ChatCompletionRequest) -> "ChatCompletionRequest":
        ccc = ccc.model_dump()
        if "seed" in ccc:
            ccc["random_seed"] = ccc["seed"]
            del ccc["seed"]
        if "lm_config" in ccc:
            del ccc["lm_config"]
        return ChatCompletionRequest.model_validate(ccc)

    def to_dict(self) -> Mapping[str, Any]:
        return self.model_dump(exclude_none=True)


class ChatCompletionResponse(models.ChatCompletionResponse):
    """
    https://github.com/mistralai/client-python/blob/main/src/mistralai/models/chat_completion.py#L86
    """
    @classmethod
    def from_dict(cls, m: Mapping[str, Any]) -> "ChatCompletionResponse":
        return ChatCompletionResponse.model_validate(m)

    def to_chat_completion_response(self) -> models.ChatCompletionResponse:
        return models.ChatCompletionResponse.model_validate(self.model_dump(exclude_none=True))



from pydantic import BaseModel, computed_field


class Value(BaseModel):
    @computed_field
    @property
    def type(self) -> str:
        return self.__class__.__name__


class Score(Value):
    value: float


class Evaluation(BaseModel):
    identifier: str
    value: Score

